import datetime

import dictdumper

dumper = dictdumper.PLIST('temp/test.plist')

test_1 = dict(
    foo=-1,
    bar='Hello, world!',
    boo=dict(
        foo_again=True,
        bar_again=memoryview(b'bytes'),
        boo_again=None,
    ),
)
dumper(test_1, name='test_1')

test_2 = dict(
    foo=[1, 2.0, 3],
    bar=(1.0, bytearray(b'a long long bytes'), 3.0),
    boo=dict(
        foo_again=b'bytestring',
        bar_again=datetime.datetime.today(),
        boo_again=float('-inf'),
    ),
)
dumper(test_2, name='test_2')

test_3 = dict(
    foo="string",
    bar=[
        "s1", "s2", "s3",
    ],
    boo=[
        "s4", dict(s="5", j="5"), "s6"
    ],
    far=dict(
        far_foo=["s1", "s2", "s3"],
        far_var="s4",
    ),
    biu="s5",
)
dumper(test_3, name='test_3')



# -*- coding: utf-8 -*-
"""Date & time processing utilities."""

import datetime

__all__ = ['isoformat']

try:
    from datetime import _format_time
except ImportError:
    def _format_time(hh, mm, ss, us, timespec='auto'):
        specs = {
            'hours': '{:02d}',
            'minutes': '{:02d}:{:02d}',
            'seconds': '{:02d}:{:02d}:{:02d}',
            'milliseconds': '{:02d}:{:02d}:{:02d}.{:03d}',
            'microseconds': '{:02d}:{:02d}:{:02d}.{:06d}'
        }

        if timespec == 'auto':
            # Skip trailing microseconds when us==0.
            timespec = 'microseconds' if us else 'seconds'
        elif timespec == 'milliseconds':
            us //= 1000
        try:
            fmt = specs[timespec]
        except KeyError:
            raise ValueError('Unknown timespec value')
        else:
            return fmt.format(hh, mm, ss, us)

try:
    from datetime import _format_offset
except ImportError:
    def _format_offset(off):
        s = ''
        if off is not None:
            if off.days < 0:
                sign = "-"
                off = -off
            else:
                sign = "+"
            hh, mm = divmod(off, datetime.timedelta(hours=1))
            mm, ss = divmod(mm, datetime.timedelta(minutes=1))
            s += "%s%02d:%02d" % (sign, hh, mm)
            if ss or ss.microseconds:
                s += ":%02d" % ss.seconds

                if ss.microseconds:
                    s += '.%06d' % ss.microseconds
        return s


def isoformat(date):
    """Generate ISO format date string."""
    if hasattr(date, 'isoformat'):
        return date.isoformat()
    if isinstance(date, datetime.date):
        return "%04d-%02d-%02d" % (date.year, date.month, date.day)
    if isinstance(date, datetime.time):
        s = _format_time(date.hour, date.minute, date.second,
                         date.microsecond, 'auto')
        off = date.utcoffset()
        tz = _format_offset(off)
        if tz:
            s += tz
        return s
    if isinstance(date, datetime.datetime):
        s = ("%04d-%02d-%02d%c" % (date.year, date.month, date.day, 'T') +
             _format_time(date.hour, date.minute, date.second,
                          date.microsecond, 'auto'))

        off = date.utcoffset()
        tz = _format_offset(off)
        if tz:
            s += tz

        return s
    raise NotImplementedError('unknown date object: %s' % date)



# -*- coding: utf-8 -*-
"""Unittest cases."""

import datetime
import os
import tempfile
import unittest

import dictdumper

ROOT = os.path.dirname(os.path.realpath(__file__))

test_1 = dict(
    foo=-1,
    bar='Hello, world!',
    boo=dict(
        foo_again=True,
        bar_again=memoryview(b'bytes'),
        boo_again=None,
    ),
)

test_2 = dict(
    foo=[1, 2.0, 3],
    bar=(1.0, bytearray(b'a long long bytes'), 3.0),
    boo=dict(
        foo_again=b'bytestring',
        bar_again=datetime.datetime(2020, 1, 31, 20, 15, 10, 163010),
        boo_again=float('-inf'),
    ),
)

test_3 = dict(
    foo="stringstringstringstringstringstringstringstringstringstring",
    bar=[
        "s1", False, "s3",
    ],
    boo=[
        "s4", dict(s="5", j="5"), "s6"
    ],
    far=dict(
        far_foo=["s1", "s2", "s3"],
        far_var="s4",
    ),
    biu=float('nan'),
)


class TestDictDumper(unittest.TestCase):
    """Test DictDumper."""

    def assertFile(self, first, second):
        with open(first) as file:
            text_first = file.read()
        with open(second) as file:
            text_second = file.read()
        self.assertEqual(text_first, text_second)

    def test_json(self):
        """Test JSON dumper."""
        rootdir = os.path.join(ROOT, 'json')
        with tempfile.TemporaryDirectory() as tempdir:
            dst = os.path.join(tempdir, 'test.json')

            dumper = dictdumper.JSON(dst)
            self.assertFile(dst, os.path.join(rootdir, 'test_0.json'))

            dumper(test_1, name='test_1')
            self.assertFile(dst, os.path.join(rootdir, 'test_1.json'))

            dumper(test_2, name='test_2')
            self.assertFile(dst, os.path.join(rootdir, 'test_2.json'))

            dumper(test_3, name='test_3')
            self.assertFile(dst, os.path.join(rootdir, 'test_3.json'))

    def test_plist(self):
        """Test PLIST dumper."""
        rootdir = os.path.join(ROOT, 'plist')
        with tempfile.TemporaryDirectory() as tempdir:
            dst = os.path.join(tempdir, 'test.plist')

            dumper = dictdumper.PLIST(dst)
            self.assertFile(dst, os.path.join(rootdir, 'test_0.plist'))

            dumper(test_1, name='test_1')
            self.assertFile(dst, os.path.join(rootdir, 'test_1.plist'))

            dumper(test_2, name='test_2')
            self.assertFile(dst, os.path.join(rootdir, 'test_2.plist'))

            dumper(test_3, name='test_3')
            self.assertFile(dst, os.path.join(rootdir, 'test_3.plist'))

    def test_tree(self):
        """Test text tree dumper."""
        rootdir = os.path.join(ROOT, 'tree')
        with tempfile.TemporaryDirectory() as tempdir:
            dst = os.path.join(tempdir, 'test.txt')

            dumper = dictdumper.Tree(dst)
            self.assertFile(dst, os.path.join(rootdir, 'test_0.txt'))

            dumper(test_1, name='test_1')
            self.assertFile(dst, os.path.join(rootdir, 'test_1.txt'))

            dumper(test_2, name='test_2')
            self.assertFile(dst, os.path.join(rootdir, 'test_2.txt'))

            dumper(test_3, name='test_3')
            self.assertFile(dst, os.path.join(rootdir, 'test_3.txt'))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""base dumper

``dictdumper.dumper`` contains ``Dumper`` only, which is an
abstract base class for all dumpers, eg. HTML, JSON, PLIST,
Tree, and XML.

"""
# Abstract Base Class of Dumpers
# Pre-define useful arguments and methods of dumpers

import abc
import os
import warnings

__all__ = ['Dumper']


def deprecated(cls):
    """Deprecation warning."""
    warnings.warn('%s is deprecated' % cls.__name__, DeprecationWarning, stacklevel=2)
    return cls


class DumperError(TypeError):
    """Unsupported content type."""


class Dumper:  # pylint: disable=metaclass-assignment
    """Abstract base class of all dumpers.

    Usage:
        >>> dumper = Dumper(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, file format of current dumper
        * filename - str, output file name

    Methods:
        * make_object - create an object with convertion information
        * object_hook - convert content for function call
        * default - check content type for function call

    Attributes:
        * _file - str, output file name
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hsrt - str, _HEADER_START
        * _hend - str, _HEADER_END

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _encode_func - check content type for function call
        * _encode_value - convert content for function call
        * _append_value - call this function to write contents

    """
    __metaclass__ = abc.ABCMeta

    ##########################################################################
    # Properties.
    ##########################################################################

    # file format of current dumper
    @property
    @abc.abstractmethod
    def kind(self):
        """File format of current dumper."""

    @property
    def filename(self):
        """Output file name."""
        return self._file

    ##########################################################################
    # Type codes.
    ##########################################################################

    __type__ = dict()

    ##########################################################################
    # Methods.
    ##########################################################################

    @staticmethod
    def make_object(o, value, **kwargs):
        """Create an object with convertion information."""
        obj = dict(
            type=type(o).__name__,
            value=value,
        )
        obj.update(kwargs)
        return obj

    def object_hook(self, o):  # pylint: disable=unused-argument,no-self-use
        """Convert content for function call."""
        return o

    def default(self, o):  # pylint: disable=unused-argument,no-self-use
        """Check content type for function call."""
        raise DumperError('unsupported content type: %s' % type(o).__name__)

    ##########################################################################
    # Attributes.
    ##########################################################################

    _sptr = os.SEEK_SET    # seek pointer
    _tctr = 1              # counter for tab level

    _hsrt = ''
    _hend = ''

    ##########################################################################
    # Data models.
    ##########################################################################

    def __new__(cls, fname, **kwargs):  # pylint: disable=unused-argument
        self = super().__new__(cls)
        return self

    def __init__(self, fname, **kwargs):  # pylint: disable=unused-argument
        self._file = fname          # dump file name
        self._dump_header()         # initialise output file

    def __call__(self, value, name=None):
        with open(self._file, 'r+') as file:
            self._append_value(value, file, name)
            self._sptr = file.tell()
            file.write(self._hend)
        return self

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _dump_header(self):
        """Initially dump file heads and tails."""
        with open(self._file, 'w') as file:
            file.write(self._hsrt)
            self._sptr = file.tell()
            file.write(self._hend)

    def _encode_func(self, o):
        """Check content type for function call."""
        name = self.__type__.get(type(o))
        if name is None:
            name = self.default(o)  # pylint: disable=assignment-from-no-return

        func = '_append_%s' % name
        return getattr(self, func)

    def _encode_value(self, o):
        """Convert content for function call."""
        return self.object_hook(o)

    @abc.abstractmethod
    def _append_value(self, value, file, name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * file - FileIO, output file
            * name - str, name of current content dict

        """



# -*- coding: utf-8 -*-
"""stream formatted output dumper

``dictdumper`` is an open source Python program works as a
stream formatted output dumper. Currently, it supports
following formats --

-  ``dictdumper.Dumper``
        abstract base class of all dumpers

-  ``dictdumper.JSON``
        dump JavaScript object notation (``JSON``)
        format file

-  ``dictdumper.PLIST``
        dump Apple property list (``PLIST``) format file

-  ``dictdumper.Tree``
        dump tree-view text (``TXT``) format file

-  ``dictdumper.XML``
        dump extensible markup language (``XML``) file;
        this is a abstract base class

-  ``dictdumper.VueJS``
        dump JavaScript file using ``Vue.js`` framework;
        this class is deprecated due to grammar error

"""
# Base Class for DictDumper
from dictdumper.dumper import Dumper  # pylint: disable=unused-import
from dictdumper.xml import XML  # pylint: disable=unused-import

# Utility Classes
from dictdumper.json import JSON
from dictdumper.plist import PLIST
from dictdumper.tree import Tree

# Deprecated Classes
from dictdumper.vuejs import VueJS  # pylint: disable=unused-import

__all__ = ['JSON', 'PLIST', 'Tree']



# -*- coding: utf-8 -*-

import datetime
import os

import dictdumper

ROOT = os.path.dirname(os.path.realpath(__file__))
dumper_0 = dictdumper.JSON(os.path.join(ROOT, 'json', 'test_0.json'))

test_1 = dict(
    foo=-1,
    bar='Hello, world!',
    boo=dict(
        foo_again=True,
        bar_again=memoryview(b'bytes'),
        boo_again=None,
    ),
)
dumper_1 = dictdumper.JSON(os.path.join(ROOT, 'json', 'test_1.json'))
dumper_1(test_1, 'test_1')

test_2 = dict(
    foo=[1, 2.0, 3],
    bar=(1.0, bytearray(b'a long long bytes'), 3.0),
    boo=dict(
        foo_again=b'bytestring',
        bar_again=datetime.datetime(2020, 1, 31, 20, 15, 10, 163010),
        boo_again=float('-inf'),
    ),
)
dumper_2 = dictdumper.JSON(os.path.join(ROOT, 'json', 'test_2.json'))
dumper_2(test_1, 'test_1')
dumper_2(test_2, 'test_2')

test_3 = dict(
    foo="stringstringstringstringstringstringstringstringstringstring",
    bar=[
        "s1", False, "s3",
    ],
    boo=[
        "s4", dict(s="5", j="5"), "s6"
    ],
    far=dict(
        far_foo=["s1", "s2", "s3"],
        far_var="s4",
    ),
    biu=float('nan'),
)
dumper_3 = dictdumper.JSON(os.path.join(ROOT, 'json', 'test_3.json'))
dumper_3(test_1, 'test_1')
dumper_3(test_2, 'test_2')
dumper_3(test_3, 'test_3')



@contextlib.contextmanager
def fileobj(o):
    if isinstance(o, str):
        try:
            sio = io.StringIO(o)
            yield sio
        finally:
            sio.close()

    # fileobj
    try:
        yield o
    finally:
        pass

with fileobj(obj) as file:
    pass



# -*- coding: utf-8 -*-
"""dumper a JSON file

``dictdumper.json`` contains ``JSON`` only, which dumpers a
JavaScript object notation (JSON) file. Usage sample is
described as below.

    >>> dumper = JSON(file_name)
    >>> dumper(content_dict_1, name=content_name_1)
    >>> dumper(content_dict_2, name=content_name_2)
    ............

"""
# Writer for JSON files
# Dump a JSON file for PCAP analyser

import collections
import datetime
import math
import os

from dictdumper._dateutil import isoformat
from dictdumper.dumper import Dumper

__all__ = ['JSON']

# head
_HEADER_START = '{\n'

# tail
_HEADER_END = '\n}'


class JSON(Dumper):
    """Dump JavaScript object notation (JSON) format file.

    Usage:
        >>> dumper = JSON(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, return 'json'
        * filename - str, output file name

    Methods:
        * make_object - create an object with convertion information
        * object_hook - convert content for function call
        * default - check content type for function call

    Attributes:
        * _file - str, output file name
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END
        * _vctr - dict, value counter dict

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _encode_func - check content type for function call
        * _encode_value - convert content for function call
        * _append_value - call this function to write contents

    Terminology:
        object    ::=  "{}" | ("{" members "}")
        members   ::=  pair | (pair "," members)
        pair      ::=  string ":" value
        array     ::=  "[]" | ("[" elements "]")
        elements  ::=  value | (value "," elements)
        value     ::=  string | number | object
                         | array | true | false | null

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'json'

    ##########################################################################
    # Type codes.
    ##########################################################################

    __type__ = {
        # string
        str: 'string',

        # date
        datetime.date: 'date',
        datetime.datetime: 'date',
        datetime.time: 'date',

        # number
        int: 'number',
        float: 'number',

        # object
        dict: 'object',

        # array
        list: 'array',

        # bool
        bool: 'bool',

        # null
        type(None): 'null',
    }

    ##########################################################################
    # Attributes.
    ##########################################################################

    _hsrt = _HEADER_START
    _hend = _HEADER_END

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, fname, **kwargs):
        super().__init__(fname, **kwargs)

        self._vctr = collections.defaultdict(int)  # value counter dict

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _encode_value(self, o):  # pylint: disable=unused-argument
        """Convert content for function call."""
        if isinstance(o, (bytes, bytearray)):
            return self.make_object(o, o.decode(errors='replace'), hex=o.hex())
        if isinstance(o, memoryview):
            tobytes = o.tobytes()
            return self.make_object(o, tobytes.decode(errors='replace'), hex=tobytes.hex())
        if isinstance(o, (tuple, set, frozenset)):
            return self.make_object(o, list(o))
        return o

    def _append_value(self, value, file, name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * file - FileIO, output file
            * name - str, name of current content dict

        """
        tabs = '\t' * self._tctr
        cmma = ',\n' if self._vctr[self._tctr] else ''
        keys = '{cmma}{tabs}"{name}": '.format(cmma=cmma, tabs=tabs, name=name)

        file.seek(self._sptr, os.SEEK_SET)
        file.write(keys)

        self._vctr[self._tctr] += 1
        self._append_object(value, file)

    ##########################################################################
    # Functions.
    ##########################################################################

    def _append_object(self, value, file):
        """Call this function to write object contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * file - FileIO, output file

        """
        labs = '{'
        file.write(labs)
        self._tctr += 1

        for (item, text) in value.items():
            tabs = '\t' * self._tctr
            cmma = ',' if self._vctr[self._tctr] else ''
            keys = '{cmma}\n{tabs}"{item}": '.format(cmma=cmma, tabs=tabs, item=item)
            file.write(keys)

            self._vctr[self._tctr] += 1

            enc_text = self._encode_value(text)
            func = self._encode_func(enc_text)
            func(enc_text, file)

        self._vctr[self._tctr] = 0
        self._tctr -= 1
        tabs = '\t' * self._tctr
        labs = '\n{tabs}{}'.format('}', tabs=tabs)
        file.write(labs)

    def _append_array(self, value, file):
        """Call this function to write array contents.

        Keyword arguments:
            * value - list, content to be dumped
            * file - FileIO, output file

        """
        val_list = [self._encode_value(item) for item in value]
        mul_line = False
        for item in val_list:
            if isinstance(item, dict):
                mul_line = True
                break

        labs = '[\n' if mul_line else '[ '
        file.write(labs)

        self._tctr += 1

        tabs = '\t' * self._tctr
        for item in val_list:
            if self._vctr[self._tctr]:
                file.write(',\n' if mul_line else ', ')
            if mul_line:
                file.write(tabs)

            self._vctr[self._tctr] += 1

            func = self._encode_func(item)
            func(item, file)

        self._vctr[self._tctr] = 0
        self._tctr -= 1

        if mul_line:
            tabs = '\t' * self._tctr
            labs = '\n{tabs}]'.format(tabs=tabs)
        else:
            labs = ' ]'
        file.write(labs)

    def _append_string(self, value, file):  # pylint: disable=no-self-use
        """Call this function to write string contents.

        Keyword arguments:
            * value - str, content to be dumped
            * file - FileIO, output file

        """
        text = str(value).replace('"', '\\"')
        labs = '"{text}"'.format(text=text)
        file.write(labs)

    def _append_date(self, value, file):  # pylint: disable=no-self-use
        """Call this function to write date contents.

        Keyword arguments:
            * value - Union[date, datetime, time], content to be dumped
            * file - FileIO, output file

        """
        text = isoformat(value)
        labs = '"{text}"'.format(text=text)
        file.write(labs)

    def _append_number(self, value, file):
        """Call this function to write number contents.

        Keyword arguments:
            * value - Union[int, float], content to be dumped
            * file - FileIO, output file

        """
        if math.isnan(value):
            text = self.make_object(value, None, number=str(value).replace('nan', 'NaN'))
            self._append_object(text, file)
        elif math.isinf(value):
            text = self.make_object(value, None, number=str(value).replace('inf', 'Infinity'))
            self._append_object(text, file)
        else:
            labs = str(value)
            file.write(labs)

    def _append_bool(self, value, file):  # pylint: disable=no-self-use
        """Call this function to write bool contents.

        Keyword arguments:
            * value - bool, content to be dumped
            * file - FileIO, output file

        """
        labs = 'true' if value else 'false'
        file.write(labs)

    def _append_null(self, value, file):  # pylint: disable=unused-argument,no-self-use
        """Call this function to write null contents.

        Keyword arguments:
            * value - NoneType, content to be dumped
            * file - FileIO, output file

        """
        labs = 'null'
        file.write(labs)



# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# README
with open('README.md', 'rb') as file:
    long_desc = file.read().decode('utf-8')

# version string
__version__ = '0.8.0'

# set-up script for pip distribution
setup(
    name='dictdumper',
    version=__version__,
    author='Jarry Shaw',
    author_email='jarryshaw@icloud.com',
    url='https://github.com/JarryShaw/dictdumper#dictdumper',
    license='Mozilla Public License 2.0 (MPL 2.0)',
    keywords='formatting dumper stream',
    description='Python dict formatted dumper.',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    python_requires='>=2.7',
    packages=['dictdumper'],
    package_data={
        '': [
            'LICENSE',
            'README.md',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: System :: Networking',
        'Topic :: Utilities',
    ]
)



# -*- coding: utf-8 -*-
"""dumper a tree-view file

``dictdumper.tree`` contains ``Tree`` only, which dumpers a
tree-view text (TXT) format file. Usage sample is described
as below.

    >>> dumper = Tree(filename)
    >>> dumper(content_dict_1, name=contentname_1)
    >>> dumper(content_dict_2, name=contentname_2)
    ............

"""
# Writer for treeview text files
# Dump a TEXT file for PCAP analyser

import contextlib
import datetime
import math
import os
import textwrap

from dictdumper._dateutil import isoformat
from dictdumper.dumper import Dumper

__all__ = ['Tree']

# headers
_HEADER_START = ''  # head
_HEADER_END = ''    # tail

# templates
_TEMP_BRANCH = '  |   '  # branch
_TEMP_SPACES = '      '  # space


@contextlib.contextmanager
def indent(ctx, branch=True):
    """Indentation context."""
    if branch:
        ctx.append(_TEMP_BRANCH)
    else:
        ctx.append(_TEMP_SPACES)
    yield
    ctx.pop()


try:
    from contextlib import nullcontext
except ImportError:
    class nullcontext:
        """Context manager that does no additional processing."""
        def __enter__(self):
            return self
        def __exit__(self, *excinfo):
            pass


class Tree(Dumper):
    """Dump a tree-view text (TXT) format file.

    Usage:
        >>> dumper = Tree(filename)
        >>> dumper(content_dict_1, name=contentname_1)
        >>> dumper(content_dict_2, name=contentname_2)
        ............

    Properties:
        * kind - str, file format of current dumper
        * filename - str, output file name

    Methods:
        * make_object - create an object with convertion information
        * object_hook - convert content for function call
        * default - check content type for function call

    Attributes:
        * _file - str, output file name
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END
        * _bctx - list, blank branch context record
        * _nctr - int, branch number counter

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _encode_func - check content type for function call
        * _encode_value - convert content for function call
        * _append_value - call this function to write contents

    Terminology:
        value   ::=  branch | array | string | number | bool | N/A
        string
          |-- string
          |     |-- string -> value
          |     |-- string
          |     |     |-- string -> value
          |     |     |-- string -> value
          |     |-- string -> value
          |     |-- string -> value
          |           |-- string -> value
          |           |-- string -> value
          |-- string -> value, value, value
          |-- string -> True
          |-- string -> False
          |-- string -> N/A
          |-- string -> value
          |-- string -> value

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'txt'

    ##########################################################################
    # Type codes.
    ##########################################################################

    __type__ = {
        # string
        str: 'string',

        # bool
        bool: 'bool',

        # branch
        dict: 'branch',

        # none
        type(None): 'none',

        # date
        datetime.date: 'date',
        datetime.datetime: 'date',
        datetime.time: 'date',

        # number
        int: 'number',
        float: 'number',
        complex: 'number',

        # bytes
        bytes: 'bytes',

        # array
        list: 'array',
    }

    ##########################################################################
    # Methods.
    ##########################################################################

    @staticmethod
    def check_newline(value):
        """Check if newline is needed."""
        if isinstance(value, dict):
            return True
        if isinstance(value, str):
            return len(value) > 2
        if isinstance(value, bytes):
            return len(value.hex()) > 32
        return False

    ##########################################################################
    # Attributes.
    ##########################################################################

    _nctr = 0
    _bctx = ''

    _hsrt = _HEADER_START
    _hend = _HEADER_END

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _encode_value(self, o):  # pylint: disable=unused-argument
        """Convert content for function call."""
        if isinstance(o, bytearray):
            return self.make_object(o, bytes(o), text=o.decode(errors='replace'))
        if isinstance(o, memoryview):
            tobytes = o.tobytes()
            return self.make_object(o, tobytes, text=tobytes.decode(errors='replace'))
        if isinstance(o, (tuple, set, frozenset)):
            return self.make_object(o, list(o))
        return o

    def _append_value(self, value, file, name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * file - FileIO, output file
            * name - str, name of current content dict

        """
        file.seek(self._sptr, os.SEEK_SET)
        if self._nctr > 0:
            file.write('\n')
        file.write(name)

        self._bctx = list()  # blank branch indent context
        self._append_branch(value, file)

        self._nctr += 1
        file.write('\n')

    ##########################################################################
    # Functions.
    ##########################################################################

    def _append_branch(self, value, file):  # pylint: disable=inconsistent-return-statements
        """Call this function to write branch contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * file - FileIO, output file

        """
        if not value:
            file.write(' ')
            return self._append_none(None, file)

        vlen = len(value)
        for (vctr, (item, text)) in enumerate(value.items(), start=1):
            file.write('\n' + ''.join(self._bctx))
            file.write('  |-- {item} '.format(item=item))

            with indent(self._bctx, branch=vctr != vlen):
                enc_text = self._encode_value(text)
                func = self._encode_func(enc_text)
                func(enc_text, file)

    def _append_array(self, value, file):  # pylint: disable=inconsistent-return-statements
        """Call this function to write array contents.

        Keyword arguments:
            * value - list, content to be dumped
            * file - FileIO, output file

        """
        if not value:
            file.write(' ')
            return self._append_none(None, file)

        vlen = len(value)
        for (vctr, item) in enumerate(value, start=1):
            file.write('\n' + ''.join(self._bctx) + '  |-')

            if vctr != vlen:
                ctx = indent(self._bctx)
            else:
                ctx = nullcontext()

            with ctx:
                enc_text = self._encode_value(item)

                if self.check_newline(enc_text):
                    file.write('-> --')

                func = self._encode_func(enc_text)
                func(enc_text, file)

    def _append_string(self, value, file):  # pylint: disable=inconsistent-return-statements
        """Call this function to write string contents.

        Keyword arguments:
            * value - str, content to be dumped
            * file - FileIO, output file

        """
        if not value:
            file.write(' ')
            return self._append_none(None, file)

        if len(value) <= 40:
            labs = '-> {text}'.format(text=value)
        else:
            labs = '\n' + ''.join(self._bctx) + '  |-'

            text_list = textwrap.wrap(value, 40)
            labs += '-> {text}'.format(text=text_list[0])
            for text in text_list[1:]:
                labs += '\n' + ''.join(self._bctx) + '       {text}'.format(text=text)
        file.write(labs)

    def _append_bytes(self, value, file):  # pylint: disable=inconsistent-return-statements
        """Call this function to write bytes contents.

        Keyword arguments:
            * value - bytes, content to be dumped
            * file - FileIO, output file

        """
        if not value:
            file.write(' ')
            return self._append_none(None, file)

        if len(value.hex()) <= 32:
            text = ' '.join(textwrap.wrap(value.hex(), 2))
            labs = '-> {text}'.format(text=text)
        else:
            labs = '\n' + ''.join(self._bctx) + '  |-'

            text_list = textwrap.wrap(value.hex(), 32)
            text = ' '.join(textwrap.wrap(text_list[0], 2))
            labs += '-> {text}'.format(text=text)
            for item in text_list[1:]:
                text = ' '.join(textwrap.wrap(item, 2))
                labs += '\n' + ''.join(self._bctx) + '       {text}'.format(text=text)
        file.write(labs)

    def _append_date(self, value, file):  # pylint: disable=no-self-use
        """Call this function to write date contents.

        Keyword arguments:
            * value - Union[datetime, date, time], content to be dumped
            * file - FileIO, output file

        """
        text = isoformat(value)
        labs = '-> {text}'.format(text=text)
        file.write(labs)

    def _append_number(self, value, file):  # pylint: disable=no-self-use
        """Call this function to write number contents.

        Keyword arguments:
            * value - Union[int, float, complex], content to be dumped
            * file - FileIO, output file

        """
        if math.isnan(value):
            text = str(value).replace('nan', 'NaN')
        elif math.isinf(value):
            text = str(value).replace('inf', 'Infinity')
        else:
            text = value
        labs = '-> {text}'.format(text=text)
        file.write(labs)

    def _append_bool(self, value, file):  # pylint: disable=no-self-use
        """Call this function to write bool contents.

        Keyword arguments:
            * value - bool, content to be dumped
            * file - FileIO, output file

        """
        text = 'True' if value else 'False'
        labs = '-> {text}'.format(text=text)
        file.write(labs)

    def _append_none(self, value, file):  # pylint: disable=unused-argument,no-self-use
        """Call this function to write none contents.

        Keyword arguments:
            * value - NoneType, content to be dumped
            * file - FileIO, output file

        """
        text = 'NIL'
        labs = '-> {text}'.format(text=text)
        file.write(labs)



# -*- coding: utf-8 -*-
"""dumper a PLIST file

``dictdumper.plist`` contains ``PLIST`` only, which dumpers an
Apple property list (PLIST) file. Usage sample is described
as below.

    >>> dumper = PLIST(file_name)
    >>> dumper(content_dict_1, name=content_name_1)
    >>> dumper(content_dict_2, name=content_name_2)
    ............

"""
# Dumper for PLIST files
# Write a macOS Property List file

import base64
import datetime
import os

from dictdumper.xml import XML

__all__ = ['PLIST']

# head
_HEADER_START = '''\
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
'''

# tail
_HEADER_END = '''\
</dict>
</plist>
'''


class PLIST(XML):
    """Dump Apple property list (PLIST) format file.

    Usage:
        >>> dumper = PLIST(file_name)
        >>> dumper(content_dict_1, name=content_name_1)
        >>> dumper(content_dict_2, name=content_name_2)
        ............

    Properties:
        * kind - str, return 'plist'
        * filename - str, output file name

    Methods:
        * make_object - create an object with convertion information
        * object_hook - convert content for function call
        * default - check content type for function call

    Attributes:
        * _file - str, output file name
        * _sptr - int (file pointer), indicates start of appending point
        * _tctr - int, tab level counter
        * _hrst - str, _HEADER_START
        * _hend - str, _HEADER_END

    Utilities:
        * _dump_header - initially dump file heads and tails
        * _encode_func - check content type for function call
        * _encode_value - convert content for function call
        * _append_value - call this function to write contents

    Terminology:
        value    ::=  array | dict | string | data
                        | date | integer | real | bool
        array    ::=  "<array>" value* "</array>"
        dict     ::=  "<dict>" ("<key>" str "</key>" value)* "</dict>"
        string   ::=  "<string>" str "</string>"
        data     ::=  "<data>" bytes "</data>"
        date     ::=  "<date>" datetime "</date>"
        integer  ::=  "<integer>" int "</integer>"
        real     ::=  "<real>" float "</real>"
        bool     ::=  "<true/>" | "<false/>"

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def kind(self):
        """File format of current dumper."""
        return 'plist'

    ##########################################################################
    # Type codes.
    ##########################################################################

    __type__ = {
        # string
        str: 'string',

        # bool
        bool: 'bool',

        # dict
        dict: 'dict',

        # date
        datetime.date: 'date',
        datetime.datetime: 'date',

        # integer
        int: 'integer',

        # real
        float: 'real',

        # data
        bytes: 'data',

        # array
        list: 'array',
    }

    ##########################################################################
    # Attributes.
    ##########################################################################

    _hsrt = _HEADER_START
    _hend = _HEADER_END

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _encode_value(self, o):  # pylint: disable=unused-argument
        """Convert content for function call."""
        if o is None:
            return self.make_object(o, 'None')
        if isinstance(o, bytearray):
            return self.make_object(o, bytes(o))
        if isinstance(o, memoryview):
            return self.make_object(o, o.tobytes())
        if isinstance(o, (tuple, set, frozenset)):
            return self.make_object(o, list(o))
        return o

    def _append_value(self, value, _file, _name):
        """Call this function to write contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * _file - FileIO, output file
            * _name - str, name of current content dict

        """
        _tabs = '\t' * self._tctr
        _keys = '{tabs}<key>{name}</key>\n'.format(tabs=_tabs, name=_name)
        _file.seek(self._sptr, os.SEEK_SET)
        _file.write(_keys)

        self._append_dict(value, _file)

    ##########################################################################
    # Functions.
    ##########################################################################

    def _append_dict(self, value, file):
        """Call this function to write dict contents.

        Keyword arguments:
            * value - dict, content to be dumped
            * file - FileIO, output file

        """
        tabs = '\t' * self._tctr
        labs = '{tabs}<dict>\n'.format(tabs=tabs)
        file.write(labs)
        self._tctr += 1

        for (item, text) in value.items():
            if text is None:
                continue

            tabs = '\t' * self._tctr
            keys = '{tabs}<key>{item}</key>\n'.format(tabs=tabs, item=item)
            file.write(keys)

            enc_text = self._encode_value(text)
            func = self._encode_func(enc_text)
            func(enc_text, file)

        self._tctr -= 1
        tabs = '\t' * self._tctr
        labs = '{tabs}</dict>\n'.format(tabs=tabs)
        file.write(labs)

    def _append_array(self, value, file):
        """Call this function to write array contents.

        Keyword arguments:
            * value - list, content to be dumped
            * file - FileIO, output file

        """
        tabs = '\t' * self._tctr
        labs = '{tabs}<array>\n'.format(tabs=tabs)
        file.write(labs)
        self._tctr += 1

        for item in value:
            if item is None:
                continue

            enc_text = self._encode_value(item)
            func = self._encode_func(enc_text)
            func(enc_text, file)

        self._tctr -= 1
        tabs = '\t' * self._tctr
        labs = '{tabs}</array>\n'.format(tabs=tabs)
        file.write(labs)

    def _append_string(self, value, file):
        """Call this function to write string contents.

        Keyword arguments:
            * value - str, content to be dumped
            * file - FileIO, output file

        """
        tabs = '\t' * self._tctr
        text = value
        labs = '{tabs}<string>{text}</string>\n'.format(tabs=tabs, text=text)
        file.write(labs)

    def _append_data(self, value, file):
        """Call this function to write data contents.

        Keyword arguments:
            * value - bytes, content to be dumped
            * file - FileIO, output file

        """
        # binascii.b2a_base64(value) -> plistlib.Data
        # binascii.a2b_base64(Data) -> value(bytes)

        tabs = '\t' * self._tctr
        text = base64.b64encode(value).decode()
        labs = '{tabs}<data>{text}</data>\n'.format(tabs=tabs, text=text)
        file.write(labs)

    def _append_date(self, value, file):
        """Call this function to write date contents.

        Keyword arguments:
            * value - Union[date, datetime], content to be dumped
            * file - FileIO, output file

        """
        tabs = '\t' * self._tctr
        text = value.strftime(r'%Y-%m-%dT%H:%M:%S.%fZ')
        labs = '{tabs}<date>{text}</date>\n'.format(tabs=tabs, text=text)
        file.write(labs)

    def _append_integer(self, value, file):
        """Call this function to write integer contents.

        Keyword arguments:
            * value - int, content to be dumped
            * file - FileIO, output file

        """
        tabs = '\t' * self._tctr
        text = value
        labs = '{tabs}<integer>{text}</integer>\n'.format(tabs=tabs, text=text)
        file.write(labs)

    def _append_real(self, value, file):
        """Call this function to write real contents.

        Keyword arguments:
            * value - float, content to be dumped
            * file - FileIO, output file

        """
        tabs = '\t' * self._tctr
        text = value
        labs = '{tabs}<real>{text}</real>\n'.format(tabs=tabs, text=text)
        file.write(labs)

    def _append_bool(self, value, file):
        """Call this function to write bool contents.

        Keyword arguments:
            * value - bool, content to be dumped
            * file - FileIO, output file

        """
        tabs = '\t' * self._tctr
        text = '<true/>' if value else '<false/>'
        labs = '{tabs}{text}\n'.format(tabs=tabs, text=text)
        file.write(labs)
